/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2012  All Rights Reserved
 */
package java.lang.invoke;

import static java.lang.invoke.MethodType.methodType;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.lang.invoke.MethodHandles.Lookup;
import java.lang.ref.WeakReference;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Collections;
import java.util.Map;
import java.util.WeakHashMap;

import sun.misc.Unsafe;

import com.ibm.oti.vm.VMLangAccess;

/*
 * A factory class that manages the relationship between ClassLoaders and the
 * injected SecurityFrame class, which acts as a trampoline to provide the
 * correct view to caller sensitive methods.
 * This class also provides an API to determine which methods are sensitive to
 * their callers.
 */
class SecurityFrameInjector {

	/* Map from ClassLoader to injected java.lang.invoke.SecurityFrame.  Map is required as findClass() will
	 * search the hierarchy.
	 */
	static Map<ClassLoader, WeakReference<Class<?>>> LoaderToSecurityFrameClassMap = Collections.synchronizedMap(new WeakHashMap<ClassLoader, WeakReference<Class<?>>>());

	static byte[] securityFrameClassBytes = null;

	/* Private class to act as a lock as using 'new Object()' for a lock
	 * prevents lock word optimizations.
	 */
	static class SecurityFrameInjectorLoaderLock { };
	/* Single shared lock to control attempts to define SecurityFrame in different loaders */
	private static SecurityFrameInjectorLoaderLock loaderLock = new SecurityFrameInjectorLoaderLock();

	/*
	 * Methods that do security stackwalks (getCallerClass(), getStackClass(), checkMemberAccess
	 * etc) require special handling when using MethodHandles.  This method identifies the set
	 * of methods needing special handling.
	 *
	 * Entries in this list restrict access even when there is no security manager in place
	 * and therefore even private methods that check their callers must be included in the
	 * list.  (It is possible to use Reflection + setAccessible + MHs.unreflect to get a MH
	 * to one of these methods)
	 */
	static boolean requiresWrappingForSecurityStackWalks(Class<?> clazz, String methodName) {
		try {
			if (MethodHandleUtils.isCallerSensitive(clazz, methodName)){
				return true;
			}
		} catch(NoClassDefFoundError e) {
			//suppress
		}

		switch(methodName) {
		case "getHelpers": //$NON-NLS-1$
			return com.ibm.jit.JITHelpers.class == clazz;
		case "getCrypto": //$NON-NLS-1$
			return com.ibm.jit.Crypto.class == clazz;
		case "lookup": //$NON-NLS-1$
			return Lookup.class == clazz;
		case "getStackClass": //$NON-NLS-1$
			return (Class.class == clazz) || (com.ibm.oti.vm.VM.class == clazz) || (org.apache.harmony.kernel.vm.VM.class == clazz);
		case "getStackClasses": //$NON-NLS-1$
		case "currentClassLoader": //$NON-NLS-1$
		case "currentLoadedClass": //$NON-NLS-1$
		case "forName": //$NON-NLS-1$
		case "getClassLoader": //$NON-NLS-1$
		case "getClasses": //$NON-NLS-1$
		case "getConstructor": //$NON-NLS-1$
		case "getConstructors": //$NON-NLS-1$
		case "getDeclaredClasses": //$NON-NLS-1$
		case "getDeclaredConstructor": //$NON-NLS-1$
		case "getDeclaredConstructors": //$NON-NLS-1$
		case "getDeclaredField": //$NON-NLS-1$
		case "getDeclaredFields": //$NON-NLS-1$
		case "getDeclaredMethod": //$NON-NLS-1$
		case "getDeclaredMethods": //$NON-NLS-1$
		case "getField": //$NON-NLS-1$
		case "getFields": //$NON-NLS-1$
		case "getMethod": //$NON-NLS-1$
		case "getMethods": //$NON-NLS-1$
		case "newInstance": //$NON-NLS-1$
			return Class.class == clazz;
		case "getCallerClass": //$NON-NLS-1$
			return System.class == clazz;
		case "callerClassLoader": //$NON-NLS-1$
			return ClassLoader.class == clazz || (org.apache.harmony.kernel.vm.VM.class == clazz);
		case "getStackClassLoader": //$NON-NLS-1$
			return (ClassLoader.class == clazz) || (com.ibm.oti.vm.VM.class == clazz) || (org.apache.harmony.kernel.vm.VM.class == clazz);
		case "getCallerClassLoader": //$NON-NLS-1$
		case "getParent": //$NON-NLS-1$
		case "getSystemClassLoader": //$NON-NLS-1$
			return ClassLoader.class == clazz;
		case "getNonBootstrapClassLoader": //$NON-NLS-1$
		case "caller2ClassLoader": //$NON-NLS-1$
			return com.ibm.oti.vm.VM.class == clazz;
		case "bootCallerClassLoader": //$NON-NLS-1$
			return org.apache.harmony.kernel.vm.VM.class == clazz;
		case "load": //$NON-NLS-1$
		case "loadLibrary": //$NON-NLS-1$
			return System.class == clazz;
		case "getContextClassLoader": //$NON-NLS-1$
			return Thread.class == clazz;
		case "LatestUserDefinedLoader": //$NON-NLS-1$
			return com.ibm.oti.vm.ORBVMHelpers.class == clazz;
		}
		return false;
	}

	/*
	 * Reads and caches the classfile bytes for java.lang.invoke.SecurityFrame
	 */
	static byte[] initializeSecurityFrameClassBytes() {
		if (securityFrameClassBytes == null) {
			synchronized(Lookup.class) {
				if (securityFrameClassBytes == null) {
					securityFrameClassBytes = AccessController.doPrivileged(new PrivilegedAction<byte[]>() {
						public byte[] run() {
							try {
								InputStream is = Lookup.class.getResourceAsStream("/java/lang/invoke/SecurityFrame.class"); //$NON-NLS-1$
								ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
								try{
									int byteCount = 0;
									byte[] buffer = new byte[1024];
									while ((byteCount = is.read(buffer, 0, buffer.length)) != -1) {
										byteStream.write(buffer, 0, byteCount);
									}
									return byteStream.toByteArray();
								} finally {
									byteStream.close();
									is.close();
								}
							} catch(java.io.IOException e) {
								// K0564 = Unable to read java.lang.invoke.SecurityFrame.class bytes
								throw new Error(com.ibm.oti.util.Msg.getString("K0564"), e); //$NON-NLS-1$
							}
						}
					});
				}
			}
		}
		return securityFrameClassBytes;
	}

	static Class<?> probeLoaderToSecurityFrameMap(ClassLoader loader) {
		WeakReference<Class<?>> weakRef = LoaderToSecurityFrameClassMap.get(loader);
		if (weakRef != null) {
			return weakRef.get();
		}
		return null;
	}

	/*
	 * Wraps the passed in MethodHandle with a trampoline MethodHandle.  The trampoline handle
	 * is from the same classloader & protection domain as the Lookup's originalAccessClass.
	 *
	 * Ie: Class A does lookup().in(Object.class).find.... The trampoline will be built against
	 * class A, not Object.
	 */
	static MethodHandle wrapHandleWithInjectedSecurityFrame(final Lookup lookup, MethodHandle handle) {
		initializeSecurityFrameClassBytes();

		/* preserve varargs status so that the wrapper handle is also correctly varargs/notvarargs */
		boolean isVarargs = Lookup.isVarargs(handle.modifiers) || handle.isVarargsCollector();
		final MethodHandle tempFinalHandle = handle;
		MethodType originalType = handle.type;

		try {
			Object o = AccessController.doPrivileged(new PrivilegedAction<Object>() {
				public Object run() {
					VMLangAccess vma = Lookup.getVMLangAccess();
					ClassLoader rawLoader = vma.getClassloader(lookup.originalAccessClass);

					/* Probe map to determine if java.lang.invoke.SecurityFrame has already been injected
					 * into this ClassLoader.  If it hasn't, use Unsafe to define it.  We can't use
					 * Class.forName() here as it will crawl the class heirarchy and SecurityFrame must
					 * be included in each loader / protection domain.
					 */
					Class<?> injectedSecurityFrameClass = probeLoaderToSecurityFrameMap(rawLoader);
					if (injectedSecurityFrameClass == null) {
						synchronized(loaderLock) {
							injectedSecurityFrameClass = probeLoaderToSecurityFrameMap(rawLoader);
							if (injectedSecurityFrameClass == null) {
								injectedSecurityFrameClass = Unsafe.getUnsafe().defineClass(
									"java.lang.invoke.SecurityFrame", //$NON-NLS-1$
									securityFrameClassBytes,
									0,
									securityFrameClassBytes.length,
									rawLoader,
									lookup.accessClass.getProtectionDomain());
								LoaderToSecurityFrameClassMap.put(rawLoader, new WeakReference<Class<?>>(injectedSecurityFrameClass));
							}
						}
					}

					try {
						Constructor<?> constructor = injectedSecurityFrameClass.getConstructor(MethodHandle.class);
						constructor.setAccessible(true);
						return constructor.newInstance(tempFinalHandle);
					} catch (SecurityException | ReflectiveOperationException e) {
						throw new Error(e);
					}
				}
			});
			handle = Lookup.internalPrivilegedLookup.bind(o, "invoke", methodType(Object.class, Object[].class)); //$NON-NLS-1$
			handle = handle.asType(originalType);
			if (isVarargs) {
				handle = handle.asVarargsCollector(originalType.parameterType(originalType.parameterCount() - 1));
			}
		} catch (IllegalAccessException | NoSuchMethodException e) {
			throw new Error(e);
		}
		return handle;
	}

	/*
	 * Helper method to wrap a MethodHandle with an injected SecurityFrame if required.  The resulting MH chain will be:
	 * AsType(type) -> ReceiverBoundHandle(Object[]) --invokeWArgs-> original handle
	 * VarargsCollector nature will be preserved.
	 */
	static MethodHandle wrapHandleWithInjectedSecurityFrameIfRequired(Lookup lookup, MethodHandle handle, String methodName) {
		if (SecurityFrameInjector.requiresWrappingForSecurityStackWalks(handle.defc, methodName)) {
			handle = SecurityFrameInjector.wrapHandleWithInjectedSecurityFrame(lookup, handle);
		}
		return handle;
	}

}
